@@ -23,6 +23,8 @@ Boston, MA 02111-1307, USA.  */
 
 /* Substantially rewritten for XEmacs.  */
 
+/* 7-8-00 !!#### This file needs definite Mule review. */
+
 #include <config.h>
 #include "lisp.h"
 
@@ -50,6 +52,7 @@ Boston, MA 02111-1307, USA.  */
 #include "faces.h"
 #include "frame.h"
 #include "window.h"
+#include "gutter.h"
 
 #ifdef HAVE_DRAGNDROP
 #include "dragdrop.h"
@@ -113,7 +116,7 @@ x_any_window_to_frame (struct device *d, Window wdesc)
 
   /* We used to map over all frames here and then map over all widgets
      belonging to that frame. However it turns out that this was very fragile
-     as it requires our display stuctures to be in sync _and_ that the 
+     as it requires our display structures to be in sync _and_ that the
      loop is told about every new widget somebody adds. Therefore we
      now let Xt find it for us (which does a bottom-up search which
      could even be faster) */
@@ -293,8 +296,7 @@ x_wm_hack_wm_protocols (Widget widget)
   int need_delete = 1;
   int need_focus = 1;
 
-  if (!XtIsWMShell (widget))
-    abort ();
+  assert (XtIsWMShell (widget));
 
   {
     Atom type, *atoms = 0;
@@ -344,7 +346,7 @@ x_wm_store_class_hints (Widget shell, char *frame_name)
   XSetClassHint (dpy, XtWindow (shell), &classhint);
 }
 
-#ifndef HAVE_SESSION
+#ifndef HAVE_WMCOMMAND
 static void
 x_wm_maybe_store_wm_command (struct frame *f)
 {
@@ -395,7 +397,7 @@ x_wm_maybe_move_wm_command (struct frame *f)
 
     }
 }
-#endif /* !HAVE_SESSION */
+#endif /* !HAVE_WMCOMMAND */
 
 static int
 x_frame_iconified_p (struct frame *f)
@@ -437,9 +439,9 @@ static void
 init_x_prop_symbols (void)
 {
 #define def(sym, rsrc) \
-   pure_put (sym, Qx_resource_name, build_string (rsrc))
+   Fput (sym, Qx_resource_name, build_string (rsrc))
 #define defi(sym,rsrc) \
-   def (sym, rsrc); pure_put (sym, Qintegerp, Qt)
+   def (sym, rsrc); Fput (sym, Qintegerp, Qt)
 
 #if 0 /* this interferes with things. #### fix this right */
   def (Qminibuffer, XtNminibuffer);
@@ -660,15 +662,15 @@ x_set_frame_text_value (struct frame *f, Bufbyte *value,
   for (ptr = value; *ptr; ptr++)
     if (!BYTE_ASCII_P (*ptr))
       {
-        CONST char * tmp;
+        const char * tmp;
         encoding = DEVICE_XATOM_COMPOUND_TEXT (XDEVICE (FRAME_DEVICE (f)));
-        GET_C_CHARPTR_EXT_CTEXT_DATA_ALLOCA ((CONST char *) value, tmp);
+	C_STRING_TO_EXTERNAL (value, tmp, Qctext);
         new_XtValue = (String) tmp;
         break;
       }
 #endif /* MULE */
 
-  /* ### Caching is device-independent - belongs in update_frame_title. */
+  /* #### Caching is device-independent - belongs in update_frame_title. */
   Xt_GET_VALUE (FRAME_X_SHELL_WIDGET (f), Xt_resource_name, &old_XtValue);
   if (!old_XtValue || strcmp (new_XtValue, old_XtValue))
     {
@@ -759,18 +761,20 @@ x_set_frame_properties (struct frame *f, Lisp_Object plist)
 
       if (STRINGP (prop))
 	{
-	  CONST char *extprop;
+	  const char *extprop;
 
 	  if (XSTRING_LENGTH (prop) == 0)
 	    continue;
 
-	  GET_C_STRING_CTEXT_DATA_ALLOCA (prop, extprop);
+	  LISP_STRING_TO_EXTERNAL (prop, extprop, Qctext);
 	  if (STRINGP (val))
 	    {
-	      CONST Extbyte *extval;
+	      const Extbyte *extval;
 	      Extcount extvallen;
 
-	      GET_STRING_CTEXT_DATA_ALLOCA (val, extval, extvallen);
+	      TO_EXTERNAL_FORMAT (LISP_STRING, val,
+				  ALLOCA, (extval, extvallen),
+				  Qctext);
 	      XtVaSetValues (w, XtVaTypedArg, extprop,
 			     XtRString, extval, extvallen + 1,
 			     (XtArgVal) NULL);
@@ -1011,7 +1015,7 @@ x_cde_destroy_callback (Widget widget, XtPointer clientData,
 	{
 	  for (i = 0; i < dragData->numItems; i++)
 	    {
-	      XtFree(dragData->data.buffers[i].bp);
+	      XtFree((char *) dragData->data.buffers[i].bp);
 	      if (dragData->data.buffers[i].name)
 		XtFree(dragData->data.buffers[i].name);
 	    }
@@ -1104,7 +1108,7 @@ WARNING: can only handle plain/text and file: transfers!
       unsigned int modifier = 0, state = 0;
       char *Ctext;
       int numItems = 0, textlen = 0, pos = 0;
-      struct Lisp_Event *lisp_event = XEVENT(event);
+      Lisp_Event *lisp_event = XEVENT (event);
       Lisp_Object item = Qnil;
       struct gcpro gcpro1;
 
@@ -1141,12 +1145,12 @@ WARNING: can only handle plain/text and file: transfers!
 	  x_event.xbutton.y_root = lisp_event->event.button.y;
 	}
       modifier = lisp_event->event.button.modifiers;
-      if (modifier & MOD_SHIFT)   state |= ShiftMask;
-      if (modifier & MOD_CONTROL) state |= ControlMask;
-      if (modifier & MOD_META)    state |= xd->MetaMask;
-      if (modifier & MOD_SUPER)   state |= xd->SuperMask;
-      if (modifier & MOD_HYPER)   state |= xd->HyperMask;
-      if (modifier & MOD_ALT)     state |= xd->AltMask;
+      if (modifier & XEMACS_MOD_SHIFT)   state |= ShiftMask;
+      if (modifier & XEMACS_MOD_CONTROL) state |= ControlMask;
+      if (modifier & XEMACS_MOD_META)    state |= xd->MetaMask;
+      if (modifier & XEMACS_MOD_SUPER)   state |= xd->SuperMask;
+      if (modifier & XEMACS_MOD_HYPER)   state |= xd->HyperMask;
+      if (modifier & XEMACS_MOD_ALT)     state |= xd->AltMask;
       state |= Button1Mask << (lisp_event->event.button.button-1);
 
       x_event.xbutton.state = state;
@@ -1166,7 +1170,7 @@ WARNING: can only handle plain/text and file: transfers!
 	  numItems++;
 	  item = XCDR (item);
 	}
-	
+
       if (numItems)
 	{
 	  /*
@@ -1175,7 +1179,7 @@ WARNING: can only handle plain/text and file: transfers!
 	   */
 	  Ctext = (char *)xmalloc (textlen+1);
 	  Ctext[0] = 0;
-	  
+
 	  item = dragdata;
 	  while (!NILP (item))
 	    {
@@ -1186,17 +1190,17 @@ WARNING: can only handle plain/text and file: transfers!
 		  Ctext=NULL;
 		  break;
 		}
-	      strcpy (Ctext+pos, (CONST char *)XSTRING_DATA (XCAR (item)));
+	      strcpy (Ctext+pos, (const char *)XSTRING_DATA (XCAR (item)));
 	      pos += XSTRING_LENGTH (XCAR (item)) + 1;
 	      item = XCDR (item);
 	    }
 	  Ctext[pos] = 0;
-	  
+
 	  dnd_convert_cb_rec[0].callback = x_cde_convert_callback;
 	  dnd_convert_cb_rec[0].closure  = (XtPointer) Ctext;
 	  dnd_convert_cb_rec[1].callback = NULL;
 	  dnd_convert_cb_rec[1].closure  = NULL;
-	  
+
 	  dnd_destroy_cb_rec[0].callback = x_cde_destroy_callback;
 	  dnd_destroy_cb_rec[0].closure  = (XtPointer) Ctext;
 	  dnd_destroy_cb_rec[1].callback = NULL;
@@ -1213,7 +1217,7 @@ WARNING: can only handle plain/text and file: transfers!
 	}
 
       UNGCPRO;
-	  
+
       return numItems?Qt:Qnil;
     }
 
@@ -1256,7 +1260,7 @@ x_cde_transfer_callback (Widget widget, XtPointer clientData,
 	{
 	  filePath = transferInfo->dropData->data.files[ii];
 	  hurl = dnd_url_hexify_string ((char *)filePath, "file:");
-          /* ### Mule-izing required */
+          /* #### Mule-izing required */
 	  l_data = Fcons (make_string ((Bufbyte* )hurl,
 				       strlen (hurl)),
 			  l_data);
@@ -1283,7 +1287,7 @@ x_cde_transfer_callback (Widget widget, XtPointer clientData,
 	  /* what, if the data is no text, and how can I tell it? */
 	  l_data = Fcons ( list3 ( list1 ( make_string ((Bufbyte *)"text/plain", 10) ),
 				   make_string ((Bufbyte *)"8bit", 4),
-				   make_string ((Bufbyte *)transferInfo->dropData->data.buffers[ii].bp, 
+				   make_string ((Bufbyte *)transferInfo->dropData->data.buffers[ii].bp,
 						transferInfo->dropData->data.buffers[ii].size) ),
 			   l_data );
  	}
@@ -1294,7 +1298,7 @@ x_cde_transfer_callback (Widget widget, XtPointer clientData,
     enqueue=0;
 
   /* The Problem: no button and mods from CDE... */
-  if (enqueue)  
+  if (enqueue)
     enqueue_misc_user_event_pos ( frame, Qdragdrop_drop_dispatch,
 				  Fcons (l_type, l_data),
 				  0 /* this is the button */,
@@ -1331,7 +1335,7 @@ The type defaults to DndText (4).
       char *dnd_data = NULL;
       unsigned long dnd_len = 0;
       int dnd_typ = DndText, dnd_dealloc = 0;
-      struct Lisp_Event *lisp_event = XEVENT(event);
+      Lisp_Event *lisp_event = XEVENT (event);
 
       /* only drag if this is really a press */
       if (EVENT_TYPE(lisp_event) != button_press_event)
@@ -1362,7 +1366,7 @@ The type defaults to DndText (4).
 		}
 	      len = XSTRING_LENGTH (XCAR (run)) + 1;
 	      dnd_data = (char *) xrealloc (dnd_data, dnd_len + len);
-	      strcpy (dnd_data + dnd_len - 1, (CONST char *)XSTRING_DATA (XCAR (run)));
+	      strcpy (dnd_data + dnd_len - 1, (const char *)XSTRING_DATA (XCAR (run)));
 	      dnd_len += len;
 	      run = XCDR (run);
 	    }
@@ -1376,15 +1380,13 @@ The type defaults to DndText (4).
 	  if (!STRINGP (data))
 	    return Qnil;
 
-	  /* and whats with MULE data ??? */
+	  /* and what's with MULE data ??? */
 	  dnd_data = (char *)XSTRING_DATA (data);
 	  dnd_len  = XSTRING_LENGTH (data) + 1; /* the zero */
 
 	}
 
-      /*
-       * not so cross hack that converts a emacs event back to a XEvent
-       */
+      /* not so gross hack that converts an emacs event back to a XEvent */
 
       x_event.xbutton.type = ButtonPress;
       x_event.xbutton.send_event = False;
@@ -1409,12 +1411,12 @@ The type defaults to DndText (4).
 	}
 
       modifier = lisp_event->event.button.modifiers;
-      if (modifier & MOD_SHIFT)   state |= ShiftMask;
-      if (modifier & MOD_CONTROL) state |= ControlMask;
-      if (modifier & MOD_META)    state |= xd->MetaMask;
-      if (modifier & MOD_SUPER)   state |= xd->SuperMask;
-      if (modifier & MOD_HYPER)   state |= xd->HyperMask;
-      if (modifier & MOD_ALT)     state |= xd->AltMask;
+      if (modifier & XEMACS_MOD_SHIFT)   state |= ShiftMask;
+      if (modifier & XEMACS_MOD_CONTROL) state |= ControlMask;
+      if (modifier & XEMACS_MOD_META)    state |= xd->MetaMask;
+      if (modifier & XEMACS_MOD_SUPER)   state |= xd->SuperMask;
+      if (modifier & XEMACS_MOD_HYPER)   state |= xd->HyperMask;
+      if (modifier & XEMACS_MOD_ALT)     state |= xd->AltMask;
       state |= Button1Mask << (lisp_event->event.button.button-1);
 
       x_event.xbutton.state = state;
@@ -1547,13 +1549,16 @@ x_initialize_frame_size (struct frame *f)
   {
     struct window *win = XWINDOW (f->root_window);
 
-    WINDOW_LEFT (win) = FRAME_LEFT_BORDER_END (f);
-    WINDOW_TOP (win) = FRAME_TOP_BORDER_END (f);
+    WINDOW_LEFT (win) = FRAME_LEFT_BORDER_END (f)
+      + FRAME_LEFT_GUTTER_BOUNDS (f);
+    WINDOW_TOP (win) = FRAME_TOP_BORDER_END (f)
+      + FRAME_TOP_GUTTER_BOUNDS (f);
 
     if (!NILP (f->minibuffer_window))
       {
 	win = XWINDOW (f->minibuffer_window);
-	WINDOW_LEFT (win) = FRAME_LEFT_BORDER_END (f);
+	WINDOW_LEFT (win) = FRAME_LEFT_BORDER_END (f)
+	  + FRAME_LEFT_GUTTER_BOUNDS (f);
       }
   }
 
@@ -1858,7 +1863,7 @@ x_create_widgets (struct frame *f, Lisp_Object lisp_window_id,
 #ifdef EXTERNAL_WIDGET
   Window window_id = 0;
 #endif
-  CONST char *name;
+  const char *name;
   Arg al [25];
   int ac = 0;
   Widget text, container, shell;
@@ -1869,7 +1874,7 @@ x_create_widgets (struct frame *f, Lisp_Object lisp_window_id,
 #endif
 
   if (STRINGP (f->name))
-    GET_C_STRING_CTEXT_DATA_ALLOCA (f->name, name);
+    LISP_STRING_TO_EXTERNAL (f->name, name, Qctext);
   else
     name = "emacs";
 
@@ -1898,7 +1903,7 @@ x_create_widgets (struct frame *f, Lisp_Object lisp_window_id,
       char *string;
 
       CHECK_STRING (lisp_window_id);
-      string = (char *) (XSTRING_DATA (lisp_window_id));
+      string = (char *) XSTRING_DATA (lisp_window_id);
       if (string[0] == '0' && (string[1] == 'x' || string[1] == 'X'))
 	sscanf (string+2, "%lxu", &window_id);
 #if 0
@@ -1938,9 +1943,9 @@ x_create_widgets (struct frame *f, Lisp_Object lisp_window_id,
       XtSetArg (al[ac], XtNinput, True);       ac++;
       XtSetArg (al[ac], XtNminWidthCells, 10); ac++;
       XtSetArg (al[ac], XtNminHeightCells, 1); ac++;
-      XtSetArg (al[ac], XtNvisual, visual); ac++;
-      XtSetArg (al[ac], XtNdepth, depth);   ac++;
-      XtSetArg (al[ac], XtNcolormap, cmap); ac++;
+      XtSetArg (al[ac], XtNvisual, visual);    ac++;
+      XtSetArg (al[ac], XtNdepth, depth);      ac++;
+      XtSetArg (al[ac], XtNcolormap, cmap);    ac++;
     }
 
   if (!NILP (parent))
@@ -2010,7 +2015,7 @@ x_create_widgets (struct frame *f, Lisp_Object lisp_window_id,
    though, the failure to call the popup callbacks resulted in XEmacs
    not accepting any input.  Bizarre but true.  Stupid but true.
 
-   So, in case there are any other gotches floating out there along
+   So, in case there are any other gotchas floating out there along
    the same lines I've duplicated the majority of XtPopup here.  It
    assumes no grabs and that the widget is not already popped up, both
    valid assumptions for the one place this is called from. */
@@ -2036,12 +2041,6 @@ xemacs_XtPopup (Widget widget)
   Xt_SET_VALUE (widget, XtNmappedWhenManaged, True);
 }
 
-#ifdef HAVE_CDE
-/* Does this have to be non-automatic? */
-/* hack frame to respond to dnd messages */
-static XtCallbackRec dnd_transfer_cb_rec[2];
-#endif /* HAVE_CDE */
-
 /* create the windows for the specified frame and display them.
    Note that the widgets have already been created, and any
    necessary geometry calculations have already been done. */
@@ -2083,9 +2082,9 @@ x_popup_frame (struct frame *f)
 	/* tell the window manager about us. */
 	x_wm_store_class_hints (shell_widget, XtName (frame_widget));
 
-#ifndef HAVE_SESSION
+#ifndef HAVE_WMCOMMAND
 	x_wm_maybe_store_wm_command (f);
-#endif /* HAVE_SESSION */
+#endif /* HAVE_WMCOMMAND */
 
 	x_wm_hack_wm_protocols (shell_widget);
       }
@@ -2096,7 +2095,7 @@ x_popup_frame (struct frame *f)
 
 #ifdef HACK_EDITRES
   /* Allow XEmacs to respond to EditRes requests.  See the O'Reilly Xt */
-  /* Instrinsics Programming Manual, Motif Edition, Aug 1993, Sect 14.14, */
+  /* Intrinsics Programming Manual, Motif Edition, Aug 1993, Sect 14.14, */
   /* pp. 483-493. */
   XtAddEventHandler (shell_widget,           /* the shell widget in question */
 		     (EventMask) NoEventMask,/* OR with existing mask */
@@ -2107,6 +2106,8 @@ x_popup_frame (struct frame *f)
 
 #ifdef HAVE_CDE
   {
+    XtCallbackRec dnd_transfer_cb_rec[2];
+
     dnd_transfer_cb_rec[0].callback = x_cde_transfer_callback;
     dnd_transfer_cb_rec[0].closure = (XtPointer) f;
     dnd_transfer_cb_rec[1].callback = NULL;
@@ -2116,7 +2117,7 @@ x_popup_frame (struct frame *f)
  			 DtDND_FILENAME_TRANSFER | DtDND_BUFFER_TRANSFER,
  			 XmDROP_COPY, dnd_transfer_cb_rec,
  			 DtNtextIsBuffer, True,
-			 DtNregisterChildren, True, 
+			 DtNregisterChildren, True,
 			 DtNpreserveRegistration, False,
 			 NULL);
   }
@@ -2198,9 +2199,9 @@ x_init_frame_2 (struct frame *f, Lisp_Object props)
    *   will update the frame title anyway, so nothing is lost.
    * JV:
    *   It turns out it gives problems with FVWMs name based mapping.
-   *   We'll just  need to be carefull in the modeline specs.
+   *   We'll just  need to be careful in the modeline specs.
    */
-  update_frame_title (f); 
+  update_frame_title (f);
 }
 
 static void
@@ -2212,10 +2213,10 @@ x_init_frame_3 (struct frame *f)
 }
 
 static void
-x_mark_frame (struct frame *f, void (*markobj) (Lisp_Object))
+x_mark_frame (struct frame *f)
 {
-  ((markobj) (FRAME_X_ICON_PIXMAP (f)));
-  ((markobj) (FRAME_X_ICON_PIXMAP_MASK (f)));
+  mark_object (FRAME_X_ICON_PIXMAP (f));
+  mark_object (FRAME_X_ICON_PIXMAP_MASK (f));
 }
 
 static void
@@ -2485,6 +2486,18 @@ x_lower_frame (struct frame *f)
     }
 }
 
+static void
+x_enable_frame (struct frame *f)
+{
+  XtSetSensitive (FRAME_X_SHELL_WIDGET (f), True);
+}
+
+static void
+x_disable_frame (struct frame *f)
+{
+  XtSetSensitive (FRAME_X_SHELL_WIDGET (f), False);
+}
+
 /* Change from withdrawn state to mapped state. */
 static void
 x_make_frame_visible (struct frame *f)
@@ -2641,17 +2654,23 @@ x_focus_on_frame (struct frame *f)
   XFlush (XtDisplay (shell_widget)); /* hey, I'd like to DEBUG this... */
 }
 
-/* Destroy the X window of frame S.  */
+/* Destroy the X window of frame F.  */
 static void
 x_delete_frame (struct frame *f)
 {
-  Widget w = FRAME_X_SHELL_WIDGET (f);
-  Display *dpy = XtDisplay (w);
+  Display *dpy;
 
-#ifndef HAVE_SESSION
+#ifndef HAVE_WMCOMMAND
   if (FRAME_X_TOP_LEVEL_FRAME_P (f))
     x_wm_maybe_move_wm_command (f);
-#endif /* HAVE_SESSION */
+#endif /* HAVE_WMCOMMAND */
+
+#ifdef HAVE_CDE
+  DtDndDropUnregister (FRAME_X_TEXT_WIDGET (f));
+#endif /* HAVE_CDE */
+
+  assert (FRAME_X_SHELL_WIDGET (f) != 0);
+  dpy = XtDisplay (FRAME_X_SHELL_WIDGET (f));
 
 #ifdef EXTERNAL_WIDGET
   expect_x_error (dpy);
@@ -2659,26 +2678,37 @@ x_delete_frame (struct frame *f)
      window-to-widget hierarchy maintained by Xt, we have to call
      XtUnrealizeWidget() here.  Xt can really suck. */
   if (f->being_deleted)
-    XtUnrealizeWidget (w);
-  XtDestroyWidget (w);
+    XtUnrealizeWidget (FRAME_X_SHELL_WIDGET (f));
+  XtDestroyWidget (FRAME_X_SHELL_WIDGET (f));
   x_error_occurred_p (dpy);
 #else
-  XtDestroyWidget (w);
-  XFlush (dpy);   /* make sure the windows are really gone! */
+  XtDestroyWidget (FRAME_X_SHELL_WIDGET (f));
+  /* make sure the windows are really gone! */
+  /* #### Is this REALLY necessary? */
+  XFlush (dpy);
 #endif /* EXTERNAL_WIDGET */
 
+  FRAME_X_SHELL_WIDGET (f) = 0;
+
   if (FRAME_X_GEOM_FREE_ME_PLEASE (f))
-    xfree (FRAME_X_GEOM_FREE_ME_PLEASE (f));
-  xfree (f->frame_data);
-  f->frame_data = 0;
+    {
+      xfree (FRAME_X_GEOM_FREE_ME_PLEASE (f));
+      FRAME_X_GEOM_FREE_ME_PLEASE (f) = 0;
+    }
+
+  if (f->frame_data)
+    {
+      xfree (f->frame_data);
+      f->frame_data = 0;
+    }
 }
 
 static void
 x_update_frame_external_traits (struct frame* frm, Lisp_Object name)
 {
-  Arg av[10];
+  Arg al[10];
   int ac = 0;
-  Lisp_Object frame = Qnil;
+  Lisp_Object frame;
 
   XSETFRAME(frame, frm);
 
@@ -2690,7 +2720,7 @@ x_update_frame_external_traits (struct frame* frm, Lisp_Object name)
      if (!EQ (color, Vthe_null_color_instance))
        {
 	 fgc = COLOR_INSTANCE_X_COLOR (XCOLOR_INSTANCE (color));
-	 XtSetArg (av[ac], XtNforeground, (void *) fgc.pixel); ac++;
+	 XtSetArg (al[ac], XtNforeground, (void *) fgc.pixel); ac++;
        }
    }
   else if (EQ (name, Qbackground))
@@ -2701,7 +2731,7 @@ x_update_frame_external_traits (struct frame* frm, Lisp_Object name)
      if (!EQ (color, Vthe_null_color_instance))
        {
 	 bgc = COLOR_INSTANCE_X_COLOR (XCOLOR_INSTANCE (color));
-	 XtSetArg (av[ac], XtNbackground, (void *) bgc.pixel); ac++;
+	 XtSetArg (al[ac], XtNbackground, (void *) bgc.pixel); ac++;
        }
 
      /* Really crappy way to force the modeline shadows to be
@@ -2714,14 +2744,14 @@ x_update_frame_external_traits (struct frame* frm, Lisp_Object name)
      Lisp_Object font = FACE_FONT (Vdefault_face, frame, Vcharset_ascii);
 
      if (!EQ (font, Vthe_null_font_instance))
-       XtSetArg (av[ac], XtNfont,
+       XtSetArg (al[ac], XtNfont,
 		 (void *) FONT_INSTANCE_X_FONT (XFONT_INSTANCE (font)));
      ac++;
    }
   else
    abort ();
 
-  XtSetValues (FRAME_X_TEXT_WIDGET (frm), av, ac);
+  XtSetValues (FRAME_X_TEXT_WIDGET (frm), al, ac);
 
 #ifdef HAVE_TOOLBARS
   /* Setting the background clears the entire frame area
@@ -2771,6 +2801,8 @@ console_type_create_frame_x (void)
   CONSOLE_HAS_METHOD (x, set_mouse_position);
   CONSOLE_HAS_METHOD (x, raise_frame);
   CONSOLE_HAS_METHOD (x, lower_frame);
+  CONSOLE_HAS_METHOD (x, enable_frame);
+  CONSOLE_HAS_METHOD (x, disable_frame);
   CONSOLE_HAS_METHOD (x, make_frame_visible);
   CONSOLE_HAS_METHOD (x, make_frame_invisible);
   CONSOLE_HAS_METHOD (x, iconify_frame);
